#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>

#include "fes.h"

typedef enum { enumeration, moebius } solution_enum_t;

uint64_t timeSecondStep = 0;

void enumeration_wrapper(LUT_t, int, int, pck_vector_t*, solution_callback_t, void*, int, int);

// this callback is used when there are more than 32 equations
int solution_tester(void *_state, uint64_t size, uint64_t* n_solutions) {
  wrapper_state_t * state = (wrapper_state_t *) _state;
  uint64_t start = rdtsc();

  assert( state->degree < enumerated_degree_bound); // enumerated_degree_bound is defined in fes.h
 
  uint64_t* corrects_solutions = malloc(sizeof(uint64_t)*size);
  uint64_t current_solution;
  int index_correct_solution = 0;
  int is_correct;
  int j;
  
  for(uint64_t i=0; i<size; i++){
    is_correct = 1;
    j = 0;
    current_solution = n_solutions[i];
    while(is_correct && j<(state->n_batches)){
      if (packed_eval(state->testing_LUT->LUT, state->n, state->degree, state->G[j], current_solution) != 0)
	is_correct = 0;
      j++;
    }
    if (is_correct){
      corrects_solutions[index_correct_solution] = current_solution;
      index_correct_solution++;
    }
  }

  timeSecondStep += (rdtsc() - start);

  if (index_correct_solution == 0)
    return 0;   // no new solution found, keep searching

  // report solution to the actual callback, and ask whether it wants to keep going
  return (*(state->callback))(state->callback_state, index_correct_solution, corrects_solutions);    
}

void moebius_wrapper(int n, pck_vector_t F[], solution_callback_t callback, void* callback_state, int verbose) {
  if (verbose) printf("fes: so far, only plain-C code is available for the moebius transform\n");      
  moebius_transform(n, F, callback, callback_state);
  
}

void next_set(int n, int d, int set[]) {
  if (d == 0) return;
  set[0] += 1;
  if (set[0] == n) {
    next_set(n-1, d-1, &set[1]);
    if (d > 0) set[0] = set[1] + 1;
  }
}

// ---------- convert the input given by the SAGE interface to the right format
//  assumes that F (the target array) is already allocated
void convert_input_equations(const int n, const int degree, int from, int to, int ***coeffs, idx_lut_t *idx_LUT, pck_vector_t F[]) {

  assert(to-from <= (int) (8*sizeof(pck_vector_t)));
  vector_t x = init_vector(to-from);   // this is used to pack the equations in memory words

  int set[ n ]; // represent the monomial `m` enumerated below
  for(int j=0; j<n; j++)
    set[j] = -1;

  for(int d=0; d<degree+1; d++) {   // collect degree-d terms
    for(int j=0; j<d; j++) set[j] = d-1-j;

    for(uint64_t m=0; m<binomials[n][d]; m++) { // iterates over all monomials of degree d
      // loop invariant: `set` describes the m-th monomial

      for(int e=from; e<to; e++)  // skim through all the equations
	x[e-from] = coeffs[e][d][m];
      F[ set2int( idx_LUT, set) ] = pack(to-from, x);

      next_set(n, n, &set[0]); // maintain invariant
    }
  }
  free(x);
}

// -------------------------------------

void exhaustive_search_wrapper(const int n, int n_eqs, const int degree, int ***coeffs, solution_callback_t callback, void* callback_state, int verbose, int T) {

  bool must_free_tester_state = false;
  const uint64_t N = n_monomials(n, degree);

  if (T == 0) T = ((degree == 2) ? 3 : 2);

  if (n > 64) {
    printf("fes: Your equations have more than 64 variables, but the FES library handles at most 64.\n");
    printf("fes: You may want to specialize (n-64) variables to fit into the constraints.\n");
    printf("fes: (also, do you realize that it will take a **LONG** time?)\n");
  }

  int enumerated_equations = 128 >> (T);

  // ---------- decide which method to use

  // this test is too simple. Should take `n` into account as well
  solution_enum_t solution = (degree < enumerated_degree_bound) ? enumeration : moebius;

  // --------- allocate/initialize some of our data structures
  pck_vector_t *F = NULL;
  idx_lut_t* idx_LUT = NULL; 

  if (solution == enumeration) {
    if (verbose) printf("fes: low degree (%d)--> using enumeration code\n", degree);

    idx_LUT = init_deginvlex_LUT(n, degree);    
    F = calloc(N, sizeof(pck_vector_t));
    if (F == NULL) perror("[fes/wrapper/enumeration/allocate F]");
    
  } else { //solution == moebius
    if (verbose) printf("fes: ''large'' degree --> using degree-independent moebius transform\n");    
    
    idx_LUT = init_lex_LUT(n);
    F = calloc(1 << n, sizeof(pck_vector_t));
    if (F == NULL) perror("[fes/wrapper/moebius/allocate F]");
  }
  
  assert (F != NULL);
  
  // ---------- deal where the case where there is more equations than what the kernel(s) deals with
  
  if (verbose) printf("fes: T = %d, %d enumerated equations for %d needed\n", T, enumerated_equations, n_eqs);
  if (n_eqs <= enumerated_equations)
    // this is the simple case where the enumeration itself is enough
    // prepare the input for the enumeration
    convert_input_equations(n, degree, 0, n_eqs, coeffs, idx_LUT, F); 

   else {
     // if there are more equations that what we can enumerate simultaneously,
     // we just deal with the first `enumerated_equations`, and then check 
     // any eventual solutions of these against the remaining equations

     if (verbose) printf("fes: too many equations. Wrapping tester around enumerative kernel.\n");

     // we split the equations into "batches" of `enumerated_equations` each
     int n_batches = n_eqs / enumerated_equations;
     if ( (n_eqs % enumerated_equations) > 0) n_batches++;

     // the first batch goes into the enumeration code
     convert_input_equations(n, degree, 0, enumerated_equations, coeffs, idx_LUT, F); // prepare the input for the enumeration
     
     // the next batches will be used by the tester. They must be in deginvlex order
     idx_lut_t *testing_LUT = idx_LUT;
     if (solution == moebius) testing_LUT = init_lex_LUT(n);

     pck_vector_t **G = NULL;
     G = calloc(n_batches-1, sizeof(pck_vector_t *));
     if (G == NULL) perror("[fes/wrapper/enumeration/allocate G]");
     assert(G != NULL);

     for(int i=1; i<n_batches; i++) {
       G[i-1] = calloc(N, sizeof(pck_vector_t));
       if (G[i-1] == NULL) perror("[fes/wrapper/enumeration/allocate G[i]]");
       assert(G[i-1] != NULL);
       convert_input_equations(n, degree, enumerated_equations*i, min(n_eqs, enumerated_equations*(i+1)), coeffs, testing_LUT, G[i-1]); 
     }
     
    // the "tester" needs some internal state
    wrapper_state_t * tester_state = malloc(sizeof(wrapper_state_t));
    tester_state->n = n;
    tester_state->degree = degree;
    tester_state->n_batches = n_batches-1;
    tester_state->G = G;
    tester_state->testing_LUT = testing_LUT;

    tester_state->callback = callback;
    tester_state->callback_state = callback_state;
    must_free_tester_state = true;

    callback = solution_tester;
    callback_state = (void *) tester_state;
  }

  // ------------ start actual computation
  if (verbose) fprintf(stderr, "fes: starting kernel\n");
  uint64_t start = rdtsc();

  if (degree < enumerated_degree_bound)
    enumeration_wrapper(idx_LUT->LUT, n, degree, F, callback, callback_state, verbose, T);
  else
    moebius_wrapper(n, F, callback, callback_state, verbose);

  uint64_t totalTime = rdtsc() - start;
  float sizeSecondStep = (timeSecondStep * 100.0) / totalTime;
  if (verbose) fprintf(stderr, "fes: %.2f CPU cycles/candidate solution\n", totalTime * 1.0 / (1ll << n));
  if (verbose) fprintf(stderr, "fes: Enumeration : %.2f%% , Verification : \%.2f%%\n", 100-sizeSecondStep, sizeSecondStep);

  // ----------- clean up

  if (must_free_tester_state) {
    wrapper_state_t * tester_state = callback_state;
    // DO STUFF HERE -- must free G
    if (solution == moebius) free_LUT(tester_state->testing_LUT);
    free(tester_state);
  }

  free_LUT(idx_LUT);
  free(F);
}

void enumeration_wrapper(LUT_t LUT, int n, int d, pck_vector_t F[], solution_callback_t callback, void* callback_state, int verbose, int T) {

  // TODO : this should probably also include a run-time check that SSE2 instructions are actually there
#ifdef HAVE_SSE2
  int sse2_code = 1;
#else
  int sse2_code = 0;
#endif

  if (n < SIMD_CHUNK_SIZE + 2) {
    if (verbose) printf("fes: small number of variables. Using sequential plain-C code\n");      
    sse2_code = 0;
  }
  
  if (!sse2_code) {
    if (verbose) printf("fes: no Using plain-C code.\n");
    switch (d) {
    case 2: exhaustive_ia32_deg_2(LUT, n, F, callback, callback_state, verbose); break;
    case 3: exhaustive_ia32_deg_3(LUT, n, F, callback, callback_state, verbose); break;
    case 4: exhaustive_ia32_deg_4(LUT, n, F, callback, callback_state, verbose); break;
    case 5: exhaustive_ia32_deg_5(LUT, n, F, callback, callback_state, verbose); break;
    case 6: exhaustive_ia32_deg_6(LUT, n, F, callback, callback_state, verbose); break;
    case 7: exhaustive_ia32_deg_7(LUT, n, F, callback, callback_state, verbose); break;
    case 8: exhaustive_ia32_deg_8(LUT, n, F, callback, callback_state, verbose); break;
    case 9: exhaustive_ia32_deg_9(LUT, n, F, callback, callback_state, verbose); break;
    default:
      assert(0);
    }
  } 
  else{
    if (verbose) printf("fes: Using SIMD code (sse2 instructions available)\n");
    if (T == 2){
      switch (d) {
#if defined(HAVE_SSE2) && defined(HAVE_64_BITS)
      case 2: exhaustive_sse2_deg_2_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 3: exhaustive_sse2_deg_3_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 4: exhaustive_sse2_deg_4_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 5: exhaustive_sse2_deg_5_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 6: exhaustive_sse2_deg_6_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 7: exhaustive_sse2_deg_7_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 8: exhaustive_sse2_deg_8_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 9: exhaustive_sse2_deg_9_T_2_el_0(LUT, n, F, callback, callback_state, verbose); break;
#endif
      default:
	assert(0);
      }
    }
    else if (T == 3){
      switch (d) {
#if defined(HAVE_SSE2) && defined(HAVE_64_BITS)
      case 2: exhaustive_sse2_deg_2_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 3: exhaustive_sse2_deg_3_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 4: exhaustive_sse2_deg_4_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 5: exhaustive_sse2_deg_5_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 6: exhaustive_sse2_deg_6_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 7: exhaustive_sse2_deg_7_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 8: exhaustive_sse2_deg_8_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 9: exhaustive_sse2_deg_9_T_3_el_0(LUT, n, F, callback, callback_state, verbose); break;
#endif
      default:
	assert(0);
      }
    }
    else if (T == 4){
      switch (d) {
#if defined(HAVE_SSE2) && defined(HAVE_64_BITS)
      case 2: exhaustive_sse2_deg_2_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 3: exhaustive_sse2_deg_3_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 4: exhaustive_sse2_deg_4_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 5: exhaustive_sse2_deg_5_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 6: exhaustive_sse2_deg_6_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 7: exhaustive_sse2_deg_7_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 8: exhaustive_sse2_deg_8_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
      case 9: exhaustive_sse2_deg_9_T_4_el_0(LUT, n, F, callback, callback_state, verbose); break;
#endif
      default:
	assert(0);
      }
    }
  }
}
